//  Copyright (C) 2003 Koji Tsukada and Satoru Takabayashi
//      All rights reserved.
//      This is free software with ABSOLUTELY NO WARRANTY.
//
//  You can redistribute it and/or modify it under the terms of
//  the GNU General Public License version 2.

using System;
using System.Runtime.InteropServices;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;

namespace FaceIcon
{
	#region Class IconFileHeader
	class IconFileHeader
	{
		public short reserved1;		
		public short type;
		public short imageCount;	//rbg}bvێ邩
		//image directory entry
		public byte width;
		public byte height;
		public byte colorDepth;
		public byte reserved2;
		public short planes;
		public short bitsPerPixel;
		public int size;
		public int offset;
		
		public IconFileHeader(ImageSize imageSize)
		{
			this.reserved1 = 0x0000;
			this.type = 0x0001;
			this.imageCount = 0x0001;
			this.width = (byte)imageSize.width;		
			this.height = (byte)imageSize.height;	
			this.colorDepth = 0x00;					
			this.reserved2 = 0x00;
			this.planes = 0x00;					
			this.bitsPerPixel = 0x00;				

			this.size = 40 + 1024 + imageSize.indexSize + imageSize.maskSize; 
			this.offset = 22 ;	//iconFileHeader̃TCY
		}

		//oCgɕϊ
		public byte [] ToBytes()
		{

			byte [] bytes = new byte[22];
			int [] byteOffsets = new int[]{0,2,4,6,7,8,9,10,12,14,18,22};
			int [] values =new int[]{this.reserved1,this.type,this.imageCount,
										this.width, this.height, this.colorDepth,
										this.reserved2, this.planes, 
										this.bitsPerPixel, this.size, this.offset};

			for(int cnt =0;cnt<byteOffsets.Length-1;cnt++)
			{
				for(int i=byteOffsets[cnt];i<byteOffsets[cnt+1];i++)
				{
					bytes[i] = BitConverter.GetBytes(values[cnt])
						[i-byteOffsets[cnt]];
				}
			}		
			return bytes;
		}

	}

	#endregion

	#region Class BitmapFileHeader
	/*
	class BitmapFileHeader
	{
		public short type;	//0,2,6,8,10,14
		public int size;
		public short reserved1;
		public short reserved2;
		public int offSets;
	}
	*/
	#endregion

	#region Class BitmapInfoHeader
	class BitmapInfoHeader
	{
		public int size;	//0,4,8,12,14,16,20,24,28,32,36,40
		public int width;
		public int height;
		public short planes;
		public short bitPerPixel;
		public int compression;
		public int sizeImage;
		public int pixelPerMeterX;
		public int pixelPerMeterY;
		public int colorPalete;
		public int colorImportant;
			
		public BitmapInfoHeader(ImageSize imageSize)
		{			
			//t@C̏ԂύX̂Lq
			this.size = 40;
			this.width = (short)imageSize.width;
			this.height= (short)imageSize.height * 2;
			this.planes = (short)1;
			this.bitPerPixel = 0x08;
			this.sizeImage = imageSize.indexSize + imageSize.maskSize;
			//indexImageDatãTCY + Mask̃TCŶ
			this.pixelPerMeterX =0;
			this.pixelPerMeterY =0; 

		}

		//oCgɕϊ
		public byte[] ToBytes()
		{
			byte [] bytes = new byte[40];
			int [] byteOffsets = new int[]{0,4,8,12,14,16,20,24,28,32,36,40};
			
			int [] values = 
				new int[]{this.size,this.width,this.height,
							 this.planes, this.bitPerPixel, this.compression,
							 this.sizeImage, this.pixelPerMeterX, 
							 this.pixelPerMeterY, this.colorPalete, 
							 this.colorImportant};

			for(int cnt =0;cnt<byteOffsets.Length-1;cnt++)
			{
				for(int i=byteOffsets[cnt];i<byteOffsets[cnt+1];i++)
				{
					bytes[i] = BitConverter.GetBytes(values[cnt])[i-byteOffsets[cnt]];
				}
			}
			return bytes;
		}
	}
	#endregion

	#region Class ColorCount
	class ColorCount
	{
		public int count;
		public short color;
	}
	#endregion

	#region Class ColorPalete
	class ColorPalete
	{
		public byte red;
		public byte green;
		public byte blue;
		public byte reserved;
	}
	#endregion
		
	#region Class ImageSize
	class ImageSize
	{
		public int width;
		public int height;
		public long dataSize;
		public int arrayWidth;
		public int indexWidth;
		public int indexSize;
		public int maskWidth;
		public int maskSize;

		public ImageSize(int width, int height)
		{
			this.width = width;
			this.height = height;
			this.arrayWidth = (int)(width * 4);
			this.indexWidth = (int)((width + 3) & 0xfffffffc);
			this.indexSize = this.indexWidth * height;
			this.maskWidth = (int)((this.indexWidth / 8 + 3) & 0xfffffffc);
			this.maskSize = this.maskWidth * height;
			return;
		}

	} 
	#endregion

	public class UtilBitmap
	{
		ImageSize imageSize;
		string iconPath;

		BitmapInfoHeader biHeader;
		IconFileHeader ifHeader;
		byte [] imageBytes;
		byte [] imageDataBytes;

		//RXgN^
		public UtilBitmap(string path, Bitmap bmp)
		{
			int imageOffset;

			this.iconPath = path;

			imageSize = new ImageSize(bmp.Width, bmp.Height);
			biHeader = new BitmapInfoHeader(imageSize);
			ifHeader = new IconFileHeader(imageSize);

			imageBytes = GetImageBytes((this.iconPath + ".tmp"),bmp);
			imageOffset =  14 + 40; 
			imageDataBytes = new byte[imageBytes.Length - imageOffset];	
			
			for(int i=0;i<imageBytes.Length - imageOffset;i++)
			{
				imageDataBytes[i] = imageBytes[i+imageOffset];
			}
			return;
		}


		public void CreateBitmapIcon()
		{
			ColorCount[] colorCounts;
			ColorPalete[] colorPaletes;
			byte [] indexImageBytes;
			byte [] maskImageBytes;

			colorCounts = new ColorCount[4096];
			colorPaletes = new ColorPalete[256];

			for(int i=0;i<4096;i++)
				colorCounts[i] = new ColorCount();

			for(int i=0;i<256;i++)
				colorPaletes[i] = new ColorPalete();


			CreateIndexPalete(colorCounts, colorPaletes);					
			indexImageBytes = CreateIndexImage (colorCounts, colorPaletes);
			maskImageBytes =  CreateMaskImage(ref indexImageBytes);
			WriteIconFile(iconPath, colorPaletes, ref indexImageBytes, ref maskImageBytes);

		}


		byte[] GetImageBytes(string tempPath, Bitmap tempBmp)
		{
			FileStream readStream;
			byte [] imageBytes;

			tempBmp.Save(tempPath, ImageFormat.Bmp);
			
			readStream = new FileStream(tempPath, FileMode.Open, FileAccess.Read);	
			imageBytes = new byte[readStream.Length];
		
			readStream.Read(imageBytes,0,(int)readStream.Length);
			readStream.Close();	
			File.Delete(iconPath + ".tmp");
			return imageBytes;
		}


		//FIndexColorPalete쐬
		void CreateIndexPalete(	ColorCount[] colorCounts, ColorPalete[] colorPaletes) 
		{
			//F
			for (int i =0;i<4096;i++)
			{
				colorCounts[i].count = 0;
				colorCounts[i].color =(short)i;
			}

			//eF4bitɕύX
			for(int i=0;i<imageSize.height;i++)
			{
				for(int j=0;j<imageSize.width;j++)
				{
					int r,g,b,index;

					b = (int)(imageDataBytes[i * imageSize.arrayWidth + j * 4] / 16);
					g = (int)(imageDataBytes[i * imageSize.arrayWidth + j * 4 + 1] / 16);
					r = (int)(imageDataBytes[i * imageSize.arrayWidth + j * 4 + 2] / 16);

					index = r*256 + g*16 + b;
					colorCounts[index].count ++;
				}
			}
			
			//pɂɎgĂF擾
			for(int i =0; i<256;i++)
			{
				for(int j = i + 1;j<4096;j++)
				{
					if(colorCounts[i].count < colorCounts[j].count)
					{
						ColorCount temp;
						temp = colorCounts[i];
						colorCounts[i] = colorCounts[j];
						colorCounts[j] = temp;
					}
				}

			}

			//\[gf[^pbgɕϊ
			for(int i=(0+1);i<(256-1);i++)
			{
				byte r,g,b;
				
				r = (byte)(((colorCounts[i].color & 0xf00) / 256)*16);// * 16
				g = (byte)(((colorCounts[i].color & 0x0f0) / 16)*16);// * 16
				b = (byte)(((colorCounts[i].color & 0x00f))*16);// * 16

				if((r == 0 && g==0 && b ==0)||(r==255 &&g==255&&b==255))
				{
					//Ȃ
				}
				else
				{
					colorPaletes[i].blue = b;
					colorPaletes[i].green = g;
					colorPaletes[i].red = r;
				}
			}

			//ŏ000, Ō111Ŋm
			colorPaletes[0].red = (byte)0;
			colorPaletes[0].green = (byte)0;
			colorPaletes[0].blue = (byte)0;
			colorPaletes[255].red = (byte)255;
			colorPaletes[255].green = (byte)255;
			colorPaletes[255].blue = (byte)255;

			return;
		}
	

		//pbg̊蓖ĂƗppxɃ\[g
		byte [] CreateIndexImage(ColorCount [] colorCounts, ColorPalete [] colorPaletes)
		{
			byte [] indexImageBytes;

			indexImageBytes = new byte[imageSize.indexSize];

			for(int i =0; i<imageSize.height;i++)
			{
				for(int j = 0;j<imageSize.width;j++)
				{
					int r1,g1,b1;
					int r2,g2,b2;
					double diff;
					byte diffNum;
					b1 = (int)(imageDataBytes[i * imageSize.arrayWidth + j * 4]);
					g1 = (int)(imageDataBytes[i * imageSize.arrayWidth + j * 4 + 1]);
					r1 = (int)(imageDataBytes[i * imageSize.arrayWidth + j * 4 + 2]);

					r2 = colorPaletes[0].red;
					g2 = colorPaletes[0].green;
					b2 = colorPaletes[0].blue;
						
					diff = Math.Pow((r2 - r1),2) + Math.Pow((g2 - g1),2) + Math.Pow((b2 - b1), 2);
					diffNum = 0;
					for(int k =1; k<256;k++)
					{
						double temp;
						r2 = colorPaletes[k].red;
						g2 = colorPaletes[k].green;
						b2 = colorPaletes[k].blue;					
							
						temp = Math.Pow((r2 - r1),2) + Math.Pow((g2 - g1),2) + Math.Pow((b2 - b1), 2);
						//܂ōŏ菬Ώ
						if(temp < diff)
						{
							diff = temp;
							diffNum = (byte)k;
						}
					}
					//Vrbg}bvf[^ɏ
					indexImageBytes[i*imageSize.indexWidth+j] = diffNum;
				}
			}
			return indexImageBytes;
		}

		
		//}XN̍쐬
		byte [] CreateMaskImage(ref byte[] indexImageBytes)
		{
			byte [] maskImageBytes = new byte[imageSize.maskSize];
			byte [] masks = new byte[] {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

			for(int i =0; i<imageSize.height;i++)
			{
				int cnt = 0,maskCnt=0;
				for(int j = 0;j<imageSize.width;j++)
				{
					if(indexImageBytes[i*imageSize.indexWidth+j] == 0x00)
					{
						maskImageBytes[i*imageSize.maskWidth + cnt]  = 
							(byte)(maskImageBytes[i*imageSize.maskWidth + cnt] | masks[maskCnt]);
					}
					maskCnt ++;
					if(maskCnt == 8) 
					{	
						cnt ++;
						maskCnt = 0;
					}
				}
			}
			return maskImageBytes;
		}
		

		//ACR`ŕۑioCg̏ój
		void WriteIconFile(string iconPath, ColorPalete [] colorPaletes, 
							ref byte[] indexImageBytes, ref byte [] maskImageBytes)
		{
			FileStream writeStream;
			byte [] ifBytes, biBytes;

			ifBytes = ifHeader.ToBytes();
			biBytes = biHeader.ToBytes();

			writeStream = new FileStream(iconPath, FileMode.Create,FileAccess.Write,FileShare.None);

			writeStream.Write(ifBytes, 0, ifBytes.Length);
			writeStream.Write(biBytes, 0, biBytes.Length);

			for(int i =0;i < 256 ;i++)
			{
				writeStream.WriteByte(colorPaletes[i].blue);
				writeStream.WriteByte(colorPaletes[i].green);
				writeStream.WriteByte(colorPaletes[i].red);
				writeStream.WriteByte(colorPaletes[i].reserved);	
			}

			writeStream.Write(indexImageBytes,0,indexImageBytes.Length);
			writeStream.Write(maskImageBytes,0,maskImageBytes.Length);
			writeStream.Close();
			return;
		}
	}


}
